import functools
import threading

import pika
import sys
import json

START_CONSUMING = 'Waiting for messages. To exit press CTRL+C'

class AsyncBaseMqWorker:
    def __init__(self, mq_username, mq_password, mq_host, mq_port, mq_virtual_host, consume_queue, produce_queue,
                 blocked_connection_timeout=None) -> None:
        self.__connection = None
        self.__channel = None
        self._mq_username = mq_username
        self._mq_password = mq_password
        self._mq_host = mq_host
        self._mq_port = mq_port
        self._mq_virtual_host = mq_virtual_host
        self._consume_queue = consume_queue
        self._produce_queue = produce_queue
        self._blocked_connection_timeout = blocked_connection_timeout
        self._threads = []

    def start_consuming(self, message: str = START_CONSUMING):
        print(message, file=sys.stderr)
        self.__channel.start_consuming()

    def ack_message(self, channel, delivery_tag):
        """Note that `channel` must be the same pika channel instance via which
        the message being ACKed was retrieved (AMQP protocol constraint).
        """
        if channel.is_open:
            print('Acknowledging message {0}'.format(delivery_tag))
            channel.basic_ack(delivery_tag)
        else:
            # Channel is already closed, so we can't ACK this message;
            # log and/or do something that makes sense for your app in this case.
            pass

    def __callback(self, connection, channel, delivery_tag, body):
        thread_id = threading.get_ident()
        message = body.decode()
        self.consume(message)

        cb = functools.partial(self.ack_message, channel, delivery_tag)
        connection.add_callback_threadsafe(cb)

        print(f'\n\r{START_CONSUMING}', file=sys.stderr)

    def consume(self, message):
        print(f'Recevied "{message}"', file=sys.stderr)

    def on_message(self, channel, method_frame, header_frame, body, args):
        (connection, threads) = args
        delivery_tag = method_frame.delivery_tag
        t = threading.Thread(target=self.__callback, args=(connection, channel, delivery_tag, body))
        t.start()
        threads.append(t)

    def create_connection(self):
        credentials = pika.PlainCredentials(
            username=self._mq_username, password=self._mq_password)
        return pika.BlockingConnection(pika.ConnectionParameters(
            host=self._mq_host, port=self._mq_port, credentials=credentials,
            heartbeat=60, virtual_host=self._mq_virtual_host,
            blocked_connection_timeout=self._blocked_connection_timeout)
        )

    def __enter__(self):
        self.__connection = self.create_connection()
        self.__channel = self.__connection.channel()

        self.__channel.queue_declare(queue=self._consume_queue, durable=True)

        self.__channel.basic_qos(prefetch_count=1)

        on_message_callback = functools.partial(self.on_message, args=(self.__connection, self._threads))
        self.__channel.basic_consume(
            queue=self._consume_queue, on_message_callback=on_message_callback)

        return self

    def publish(self, message, queue=None, persistent=True):
        connection = self.create_connection()
        produce_channel = connection.channel()
        queue = queue or self._produce_queue
        produce_channel.basic_publish(exchange='',
                                      routing_key=queue,
                                      body=json.dumps(message),
                                      properties=pika.BasicProperties(
                                          delivery_mode=pika.spec.PERSISTENT_DELIVERY_MODE, ) if persistent else None)

        connection.close()
        print(f'Published "{message}" to {queue}')

    def __exit__(self, exc_type, exc_value, traceback):
        self.close_connection()

    def close_connection(self):
        self.__channel.stop_consuming()
        # Wait for all to complete
        for thread in self._threads:
            thread.join()

        self.__connection.close()



import pika
import sys
import json

START_CONSUMING = 'Waiting for messages. To exit press CTRL+C'


class BaseMqWorker:
    def __init__(self, mq_username, mq_password, mq_host, mq_port, mq_virtual_host, consume_queue, produce_queue=None, blocked_connection_timeout=None) -> None:
        self.__connection = None
        self.__channel = None
        self._mq_username = mq_username
        self._mq_password = mq_password
        self._mq_host = mq_host
        self._mq_port = mq_port
        self._mq_virtual_host = mq_virtual_host
        self._consume_queue = consume_queue
        self._produce_queue = produce_queue
        self._blocked_connection_timeout = blocked_connection_timeout

        pass

    def start_consuming(self, message: str = START_CONSUMING):
        print(message, file=sys.stderr)
        self.__channel.start_consuming()

    def __callback(self, ch, method, props, body):
        message = body.decode()
        self.consume(message)
        ch.basic_ack(delivery_tag=method.delivery_tag)
        print(f'\n\r{START_CONSUMING}', file=sys.stderr)

    def consume(self, message):
        print(f'Recevied "{message}"', file=sys.stderr)
    
    def __enter__(self):
        credentials = pika.PlainCredentials(
            username=self._mq_username, password=self._mq_password)
        self.__connection = pika.BlockingConnection(pika.ConnectionParameters(
            host=self._mq_host, port=self._mq_port, credentials=credentials,
            heartbeat=0, virtual_host=self._mq_virtual_host, blocked_connection_timeout=self._blocked_connection_timeout)
        )

        self.__channel = self.__connection.channel()
        self.__channel.queue_declare(queue=self._consume_queue, durable=True)

        self.__channel.basic_qos(prefetch_count=1)
        self.__channel.basic_consume(
            queue=self._consume_queue, on_message_callback=self.__callback)

        return self

    
    def publish(self, message, queue=None, persistent=True):
        queue = queue or self._produce_queue

        self.__channel.basic_publish(exchange='',
                              routing_key=queue,
                              body=json.dumps(message),
                              properties=pika.BasicProperties(delivery_mode=pika.spec.PERSISTENT_DELIVERY_MODE, ) if persistent else None)
        print (f'Published "{message}" to {queue}')

    def __exit__(self, exc_type, exc_value, traceback):
        self.close_connection()

    def close_connection(self):
        self.__channel.stop_consuming()
        self.__connection.close()


from datetime import datetime, timedelta

import pytz


def get_israel_time(subtract_days=0):
    today = datetime.today() - timedelta(subtract_days)
    israel_tz = pytz.timezone('Israel')
    israel_today = today.astimezone(israel_tz).strftime('%Y-%m-%d')
    return israel_today



import base64
import json
import os
import pickle
import sys
import requests


class ModelHandler:
    @classmethod
    def send_to_models(cls, data_dict, inference_url, inference_prefix, inference_protocol, message_max_size):
        # is there a room to release memory here?
        # release memory

        # preprocessed data
        # collect metrics

        inference_results = {}
        for store in data_dict.keys():
            store_data = data_dict[store]
            df = store_data['data']
            model_url = store_data['model_url'] #serve/model_100_endpoint

            data_size = sys.getsizeof(df)
            print(f'store: {store}, data size: {data_size}')
            if data_size > message_max_size:
                raise Exception(f'error: store {store} data size > {message_max_size / (1024 * 1024)} MB')

            # send to model
            pickled = pickle.dumps(df)
            pickled_b64 = base64.b64encode(pickled)
            data_str = pickled_b64.decode('utf-8')
            url = f'{inference_protocol}://{inference_url}/{inference_prefix}/{model_url}'  # Replace with your API endpoint

            if inference_protocol == 'http':
                data = {"body": {"features": data_str}, "stores": ','.join([store])}  # Data to be sent with the POST request

                response = requests.post(url, json=data)


            else:
                raise Exception('only http is supported')

            if response.status_code == 200:
                print(f'Model inference for store {store} successful!')
                response_dict = json.loads(response.text)
                predictions = pickle.loads(base64.b64decode(response_dict['body']['predictions'].encode()))
                inference_results[store] = predictions[store]

            else :
                raise Exception(f'Model inference for store {store} failed!, status code: {response.status_code}. text: {response.text}')

        return inference_results



import json
import pickle
from io import BytesIO

import pandas as pd
from azure.storage.blob import BlobServiceClient


def exists(connection_string, container_name, blob_name):
    blob_service_client = BlobServiceClient.from_connection_string(connection_string)
    blob_client = blob_service_client.get_blob_client(container=container_name, blob=blob_name, snapshot=None)

    return blob_client.exists()


def read_blob_as_stream(connection_string, container_name, blob_name):
    blob_service_client = BlobServiceClient.from_connection_string(connection_string)
    blob_client = blob_service_client.get_blob_client(container=container_name, blob=blob_name, snapshot=None)

    download_stream = blob_client.download_blob()
    return download_stream


def read_blob_as_df(connection_string, container_name, blob_name, dtype_str=False, index_col=None, encoding='utf-8-sig'):
    download_stream = read_blob_as_stream(connection_string, container_name, blob_name)
    csv_file = download_stream.content_as_text(encoding=encoding)
    file = BytesIO(csv_file.encode())
    df = pd.read_csv(file, dtype='str' if dtype_str else None, index_col=index_col)

    return df


def read_pickle_from_blob(connection_string, container_name, blob_name):
    download_stream = read_blob_as_stream(connection_string=connection_string,
                                          container_name=container_name,
                                          blob_name=blob_name).readall()
    dictionary = pickle.loads(download_stream)
    return dictionary


def read_blob_as_json_object(connection_string, container_name, blob_name):
    download_stream = read_blob_as_stream(connection_string, container_name, blob_name).readall()
    json_data = json.loads(download_stream)

    return json_data


def save_df_to_blob(connection_string, container_name, blob_name, df: pd.DataFrame, index=False):
    """
    We assume dfs is a list of dataframes
    """

    blob_service_client = BlobServiceClient.from_connection_string(connection_string)
    blob_client = blob_service_client.get_blob_client(container=container_name, blob=blob_name)
    blob_client.upload_blob(data=df.to_csv(index=index), blob_type='BlockBlob')
    return True


def save_json_to_blob(connection_string, container_name, blob_name, dict: dict):
    stream = json.dumps(dict)

    blob_service_client = BlobServiceClient.from_connection_string(connection_string)
    blob_client = blob_service_client.get_blob_client(container=container_name, blob=blob_name)
    blob_client.upload_blob(data=stream, blob_type='BlockBlob')
    return True


def save_pickle_to_blob(connection_string, container_name, blob_name, obj: object):
    stream = pickle.dumps(obj)

    blob_service_client = BlobServiceClient.from_connection_string(connection_string)
    blob_client = blob_service_client.get_blob_client(container=container_name, blob=blob_name)
    blob_client.upload_blob(data=stream, blob_type='BlockBlob')
    return True



import csv
from io import StringIO

import pandas as pd
import psycopg2
from psycopg2 import sql
from psycopg2.extras import RealDictCursor
from psycopg2.extensions import connection, cursor


class PostgresHandler:
    def __init__(self, dbname, user, password, host, port):
        self.dbname = dbname
        self.user = user
        self.password = password
        self.host = host
        self.port = port
        self.conn: connection = None
        self.cursor: cursor = None

    def connect(self):
        self.conn = psycopg2.connect(
            dbname=self.dbname,
            user=self.user,
            password=self.password,
            host=self.host,
            port=self.port
        )
        self.cursor = self.conn.cursor(cursor_factory=RealDictCursor)

    def execute_query(self, query, params=None):
        self.cursor.execute(query, params)
        return self.cursor.fetchall()

    def read_sql_query(self, query, params=None):
        df = pd.read_sql_query(query, self.conn, params=params)
        return df

    def execute_update(self, query, params=None):
        self.cursor.execute(query, params)
        self.conn.commit()

    def insert_records(self, table_name, records):
        columns = records[0].keys()
        values = [tuple(record.values()) for record in records]

        insert_query = sql.SQL("INSERT INTO {} ({}) VALUES {}").format(
            sql.Identifier(table_name),
            sql.SQL(', ').join(map(sql.Identifier, columns)),
            sql.SQL(', ').join(sql.Placeholder() * len(records))
        )

        self.execute_update(insert_query, values)


    def delete_then_create_table_and_populate_from_df(self, table_name, df):
        from sqlalchemy import create_engine
        import io

        engine = create_engine(
            f'postgresql+psycopg2://{self.user}:{self.password}@{self.host}:{self.port}/{self.dbname}')
        df.head(0).to_sql(table_name, engine, if_exists='replace', index=False)

        conn = engine.raw_connection()
        cur = conn.cursor()
        output = io.StringIO()
        df.to_csv(output, sep='\t', header=False, index=False)
        output.seek(0)
        contents = output.getvalue()
        cur.copy_from(output, table_name, null="")  # null values become ''
        conn.commit()
        cur.close()
        conn.close()

    #TODO: this method can still be improved. for large dataframes, we can buffer loading its rows into memory
    # and then insert them in batches
    def insert_rows_from_df(self, table_name, df, null='\\\\N'):
        sio = StringIO()
        writer = csv.writer(sio, delimiter="|")
        writer.writerows(df.values)
        sio.seek(0)

        with self.conn.cursor() as c:
            c.copy_from(
                file=sio,
                table=table_name,
                columns=list(df.columns),
                sep="|",
                null=null
            )
            self.conn.commit()

    def insert_rows_from_df_schema_supported(self, table_name, schema_name, df):
        buf = StringIO()
        df.to_csv(buf, index=False)
        buf.seek(0)
        copy_sql = """
                   COPY %s FROM stdin WITH CSV HEADER
                   """
        with self.conn.cursor() as c:
            c_name = f'{schema_name}.{table_name}'
            c.copy_expert(sql=copy_sql % c_name, file=buf)
            self.conn.commit()




    def delete_records(self, table_name, condition, params=None, schema_name=None):
        if schema_name is None:
            schema_name='public'

        delete_query = sql.SQL("DELETE FROM {}.{} WHERE {}").format(
            sql.Identifier(schema_name),
            sql.Identifier(table_name),
            sql.SQL(condition)
        )
        self.execute_update(delete_query, params)



    def close(self):
        self.cursor.close()
        self.conn.close()

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

